import bs
import bsTeamGame
import bsMap
import bsUtils
import bsSpaz
import bsScoreBoard
import weakref

def getGames():
    return [DeathMatchGame]

class DeathMatchGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Death Match'

    @classmethod
    def getDescription(c,sessionType):
        return 'Kill a set number of enemies to win.'

    @classmethod
    def getSupportedSessionTypes(self):
        return ['freeForAll','teams']

    @classmethod
    def getSupportedMaps(self,sessionType):
        return bsMap.getMapList("melee")

    @classmethod
    def getSettings(self,sessionType):
        return [("Kills to Win Per Player",{'minValue':1,'default':5,'increment':1}),
                ("Time Limit",{'choices':[('None',0),('1 Minute',60),
                                        ('2 Minutes',120),('5 Minutes',300),
                                        ('10 Minutes',600),('20 Minutes',1200)],'default':0}),
                ("Respawn Times",{'choices':[('Shorter',0.25),('Short',0.5),('Normal',1.0),('Long',2.0),('Longer',4.0)],'default':1.0}),
                ("Epic Mode",{'default':False})]


    def __init__(self,session,level,settings):
        bsTeamGame.TeamGameActivity.__init__(self,session,level,settings)
        if self.settings['Epic Mode']: self._isSlowMotion = True
        self._scoreBoard = bsScoreBoard.ScoreBoard(self)

    def getInstanceDescription(self):
        return ('Crush ${ARG1} of your enemies.',self._scoreToWin)

    def getInstanceScoreBoardDescription(self):
        return ('kill ${ARG1} enemies',self._scoreToWin)

    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Epic' if self.settings['Epic Mode'] else 'ToTheDeath')

    def onTeamJoined(self,team):
        team.data['score'] = 0
        if self.hasBegun(): self._updateScoreBoard()

    def onBegin(self):
        bsTeamGame.TeamGameActivity.onBegin(self,nameSound=bs.getSound('announceDeathMatch'))
        self.setupStandardTimeLimit(self.settings['Time Limit'])
        if len(self.teams) > 0:
            self._scoreToWin = self.settings['Kills to Win Per Player'] * max(1,max(len(t.players) for t in self.teams.values()))
        else: self._scoreToWin = self.settings['Kills to Win Per Player']
        self._updateScoreBoard()
        self._dingSound = bs.getSound('dingSmall')

    def handleMessage(self,m):

        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            bsTeamGame.TeamGameActivity.handleMessage(self,m) # augment standard behavior
            self.respawnPlayer(m.spaz.getPlayer())

            player = m.spaz.getPlayer()
            killer = m.getKillerPlayer()
            if player is None or killer is None: return

            # handle team-kills
            if killer.getTeam() is player.getTeam():

                # in free-for-all, killing yourself loses you a point
                if self.getSession().getSessionType() == 'freeForAll':
                    player.getTeam().data['score'] = max(0,player.getTeam().data['score']-1)

                # in teams-mode it gives a point to the other team
                else:
                    bs.playSound(self._dingSound)
                    for team in self.teams.values():
                        if team is not killer.getTeam():
                            team.data['score'] += 1

            # killing someone on another team nets a kill
            else:
                killer.getTeam().data['score'] += 1
                bs.playSound(self._dingSound)
                # in FFA show our score since its hard to find on the scoreboard
                #if self.getSession().getSessionType() == 'freeForAll':
                try: killer.actor.setScoreText(str(killer.getTeam().data['score'])+'/'+str(self._scoreToWin),color=killer.getTeam().color,flash=True)
                except Exception: pass

            self._updateScoreBoard()

            # if someone has won, set a timer to end shortly
            # (allows the dust to clear and draws to occur if deaths are close enough)
            if any(team.data['score'] >= self._scoreToWin for team in self.teams.values()):
                self._roundEndTimer = bs.Timer(500,bs.WeakCall(self.endGame))

        else: bsTeamGame.TeamGameActivity.handleMessage(self,m)

    def _updateScoreBoard(self):
        for team in self.teams.values():
            self._scoreBoard.setTeamValue(team,team.data['score'],self._scoreToWin)

    def endGame(self):
        results = bsTeamGame.Results()
        for t in self.teams.values(): results.setTeamScore(t,t.data['score'])
        self.end(results=results)
