import bsCoopGame
import bsUtils
import bs
import bsSpaz
import bsBomb
import random
import bsAnnounce
import math
import bsUI
import bsMap
import bsScoreBoard
import weakref
import bsPowerup
import bsAchievement
import bsScoreSet

gShowedTntTip = False
gShowedCurseTip = False
gShowedLandMineTip = False

def _getDistribution(targetPoints,minDudes,maxDudes,groupCount,maxLevel):
    """ calculate a distribution of bad guys given some params """

    maxIterations = 10+maxDudes*2

    def _getTotals(groups):
        totalPoints = 0
        totalDudes = 0
        for group in groups:
            for entry in group:
                dudes = entry[1]
                totalPoints += entry[0]*dudes
                totalDudes += dudes
        return totalPoints,totalDudes

    groups = []
    for g in range(groupCount):
        groups.append([])

    types = [1]
    if maxLevel > 1: types.append(2)
    if maxLevel > 2: types.append(3)
    if maxLevel > 3: types.append(4)

    for iteration in range(maxIterations):
        # see how much we're off our target by
        totalPoints,totalDudes = _getTotals(groups)
        diff = targetPoints - totalPoints
        dudesDiff = maxDudes - totalDudes
        # add an entry if one will fit
        value = types[random.randrange(len(types))]
        group = groups[random.randrange(len(groups))]
        if len(group) == 0: maxCount = random.randint(1,6)
        else: maxCount = 2*random.randint(1,3)
        maxCount = min(maxCount,dudesDiff)
        count = min(maxCount,diff/value)
        if count > 0:
            group.append((value,count))
            totalPoints += value*count
            totalDudes += count
            diff = targetPoints - totalPoints

        totalPoints,totalDudes = _getTotals(groups)
        full = (totalPoints >= targetPoints)

        if full:
            # every so often, delete a random entry just to shake up our distribution
            if random.random() < 0.2 and iteration != maxIterations-1:
                entryCount = 0
                for group in groups:
                    for entry in group:
                        entryCount += 1
                if entryCount > 1:
                    delEntry = random.randrange(entryCount)
                    entryCount = 0
                    for group in groups:
                        for entry in group:
                            if entryCount == delEntry:
                                group.remove(entry)
                                break
                            entryCount += 1
                
            # if we don't have enough dudes, kill the group with the biggest point value
            elif totalDudes < minDudes and iteration != maxIterations-1:
                biggestValue = 9999
                biggestEntry = None
                for group in groups:
                    for entry in group:
                        if entry[0] > biggestValue or biggestEntry == None:
                            biggestValue = entry[0]
                            biggestEntry = entry
                            biggestEntryGroup = group
                if biggestEntry != None: biggestEntryGroup.remove(biggestEntry)

            # if we've got too many dudes, kill the group with the smallest point value
            elif totalDudes > maxDudes and iteration != maxIterations-1:
                smallestValue = 9999
                smallestEntry = None
                for group in groups:
                    for entry in group:
                        if entry[0] < smallestValue or smallestEntry == None:
                            smallestValue = entry[0]
                            smallestEntry = entry
                            smallestEntryGroup = group
                smallestEntryGroup.remove(smallestEntry)

            # close enough.. we're done.
            else:
                if diff == 0: break

    return groups




class OnslaughtGame(bsCoopGame.CoopGameActivity):

    tips = ['Hold any button to run.  (Trigger buttons work well if you have them)',
            'Try tricking enemies into killing eachother or running off cliffs.',
            'Try \'Cooking off\' bombs for a second or two before throwing them.',
            'It\'s easier to win with a friend or two helping.',
            'If you stay in one place, you\'re toast. Run and dodge to survive..',
            'Practice using your momentum to throw bombs more accurately.',
            'Your punches do much more damage if you are running or spinning.']

    @classmethod
    def getName(c):
        return 'Onslaught'

    @classmethod
    def getDescription(self,sessionType):
        return "Defeat all enemies."

    def __init__(self,session,settings={}):

        try: self._preset = settings['preset']
        except Exception: self._preset = 'training'

        if self._preset in ['training','pro','endless']: level = 'Doom Shroom'
        else: level = 'Courtyard'

        bsCoopGame.CoopGameActivity.__init__(self,session,level,settings)

        self._newWaveSound = bs.getSound('scoreHit01')
        self._winSound = bs.getSound("score")
        self._cashRegisterSound = bs.getSound('cashRegister')
        self._warnBeepsSound = bs.getSound("warnBeeps")
        self._aPlayerHasBeenHurt = False
        self._playerHasDroppedBomb = False

        if level == 'Doom Shroom':
            self._spawnCenter = (0,3,-5)
            self._tntSpawnPosition = (0,3,-5)
            self._powerupCenter = (0,5,-3.6)
            self._powerupSpread = (6,4)
        elif level == 'Courtyard':
            self._spawnCenter = (0,3,-2)
            self._tntSpawnPosition = (0,3,2.1)
            self._powerupCenter = (0,5,-1.6)
            self._powerupSpread = (4.6,2.7)
        else: raise Exception("Unsupported level: "+str(level))
            

    def onTransitionIn(self):

        bsCoopGame.CoopGameActivity.onTransitionIn(self)

        # show landmine tip on rookie preset
        if self._preset == 'rookie':
            global gShowedLandMineTip
            if not gShowedLandMineTip:
                gShowedLandMineTip = True
                self.tips = [{'tip':"Land-mines are a good way to stop speedy enemies.",
                              'icon':bsPowerup.getFactory().texLandMines,
                              'sound':bs.getSound('ding')}]

        # show tnt tip on pro preset
        if self._preset == 'pro':
            global gShowedTntTip
            if not gShowedTntTip:
                gShowedTntTip = True
                self.tips = [{'tip':"Take out a group of enemies by\nsetting off a bomb near a TNT box.",
                              'icon':bsBomb.getFactory().tntTex,
                              'sound':bs.getSound('ding')}]

        # curse tip on uber
        if self._preset == 'uber':
            global gShowedCurseTip
            if not gShowedCurseTip:
                gShowedCurseTip = True
                self.tips = [{'tip':"Curse boxes turn you into a ticking time bomb.\nThe only cure is to quickly grab a health-pack.",
                              'icon':bsPowerup.getFactory().texCurse,
                              'sound':bs.getSound('ding')}]

        self._spawnInfoText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                           "text",{'translate':(35,-130),
                                                   'hAttach':"left",
                                                   'vAttach':"top",
                                                   'scale':[0.55],
                                                   'color':(0.3,0.8,0.3,1.0),
                                                   'text':''},
                                           name='Spawn Info Text')
        bsUtils.playMusic('Onslaught')

        self._level = self._levelType(self,bs.Actor.UNRETAINED)
        self._scoreBoard = bsScoreBoard.ScoreBoard(self,label=bs.getResource('scoreText'))
        self._gameOver = False
        self._wave = 0
        self._canEndWave = True
        self._score = 0
        self._timeBonus = 0


    def onBegin(self):

        bsCoopGame.CoopGameActivity.onBegin(self,nameSound=bs.getSound('announceOnslaught'))
        playerCount = len(self.players)
        self._dingSound = bs.getSound('dingSmall')
        self._dingSoundHigh = bs.getSound('dingSmallHigh')

        if self._preset == 'training':

            self.createControlsHelpUI()
            self._haveTnt = False
            self._excludePowerups = ['curse','landMines']
            self._waves = [

                {'baseAngle':195,
                 'entries':([
                            {'type':bsSpaz.BomberBotLame,'spacing':5},
                            ] * playerCount)},

                {'baseAngle':130,
                 'entries':([
                            {'type':bsSpaz.ToughGuyBotLame,'spacing':5},
                            ] * playerCount)},

                {'baseAngle':195,
                 'entries':([
                            {'type':bsSpaz.BomberBotLame,'spacing':10},
                            ] * (playerCount+1))},

                {'baseAngle':130,
                 'entries':([
                            {'type':bsSpaz.ToughGuyBotLame,'spacing':10},
                            ] * (playerCount+1))},

                {'baseAngle':130,
                 'entries':([
                            {'type':bsSpaz.ToughGuyBotLame,'spacing':5} if playerCount > 1 else None,
                            {'type':bsSpaz.ToughGuyBotLame,'spacing':5},
                            {'type':None,'spacing':30},
                            {'type':bsSpaz.BomberBotLame,'spacing':5} if playerCount > 3 else None,
                            {'type':bsSpaz.BomberBotLame,'spacing':5},
                            {'type':None,'spacing':30},
                            {'type':bsSpaz.ToughGuyBotLame,'spacing':5},
                            {'type':bsSpaz.ToughGuyBotLame,'spacing':5} if playerCount > 2 else None,
                            ])},

                {'baseAngle':195,
                 'entries':([
                            {'type':bsSpaz.ChickBot,'spacing':90},
                            {'type':bsSpaz.ChickBot,'spacing':90} if playerCount > 1 else None
                            ])},
                ]

        elif self._preset == 'rookie':
            self._haveTnt = False
            self._excludePowerups = ['curse','landMines']
            self._waves = [

                {'entries':[
                        {'type':bsSpaz.NinjaBot,'point':'LeftUpperMore'} if playerCount > 2 else None,
                        {'type':bsSpaz.NinjaBot,'point':'LeftUpper'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.BomberBotStaticLame,'point':'TurretTopRight'},
                        {'type':bsSpaz.ToughGuyBotLame,'point':'RightUpper'},
                        {'type':bsSpaz.ToughGuyBotLame,'point':'RightLower'} if playerCount > 1 else None,
                        {'type':bsSpaz.BomberBotStaticLame,'point':'TurretBottomRight'} if playerCount > 2 else None,
                        ]},

                {'entries':[
                        {'type':bsSpaz.BomberBotStaticLame,'point':'TurretBottomLeft'},
                        {'type':bsSpaz.ChickBot,'point':'Left'},
                        {'type':bsSpaz.ChickBot,'point':'LeftLower'} if playerCount > 1 else None,
                        {'type':bsSpaz.ChickBot,'point':'LeftUpper'} if playerCount > 2 else None,
                        ]},
                {'entries':[
                        {'type':bsSpaz.ToughGuyBotLame,'point':'TopRight'},
                        {'type':bsSpaz.ToughGuyBot,'point':'TopHalfRight'} if playerCount > 1 else None,
                        {'type':bsSpaz.ToughGuyBotLame,'point':'TopLeft'},
                        {'type':bsSpaz.ToughGuyBotLame,'point':'TopHalfLeft'} if playerCount > 2 else None,
                        {'type':bsSpaz.ToughGuyBot,'point':'Top'},
                        {'type':bsSpaz.BomberBotStaticLame,'point':'TurretTopMiddle'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.ChickBotStatic,'point':'TurretBottomLeft'},
                        {'type':bsSpaz.ChickBotStatic,'point':'TurretBottomRight'},
                        {'type':bsSpaz.ChickBot,'point':'Bottom'},
                        {'type':bsSpaz.ChickBot,'point':'BottomHalfRight'} if playerCount > 1 else None,
                        {'type':bsSpaz.ChickBot,'point':'BottomHalfLeft'} if playerCount > 2 else None,
                        ]},

                {'entries':[
                        {'type':bsSpaz.BomberBotStaticLame,'point':'TurretTopLeft'},
                        {'type':bsSpaz.BomberBotStaticLame,'point':'TurretTopRight'},
                        {'type':bsSpaz.NinjaBot,'point':'Bottom'},
                        {'type':bsSpaz.NinjaBot,'point':'BottomHalfLeft'} if playerCount > 1 else None,
                        {'type':bsSpaz.NinjaBot,'point':'BottomHalfRight'} if playerCount > 2 else None,
                        ]},
                ]

        elif self._preset == 'pro':
            self._excludePowerups = ['curse']
            self._haveTnt = True
            self._waves = [

                {'baseAngle':-50,
                 'entries':([
                            {'type':bsSpaz.ToughGuyBot,'spacing':12} if playerCount > 3 else None,
                            {'type':bsSpaz.ToughGuyBot,'spacing':12},
                            {'type':bsSpaz.BomberBot,'spacing':6},
                            {'type':bsSpaz.BomberBot,'spacing':6},
                            {'type':bsSpaz.BomberBot,'spacing':6} if playerCount > 1 else None,
                            {'type':bsSpaz.ToughGuyBot,'spacing':12},
                            {'type':bsSpaz.ToughGuyBot,'spacing':12} if playerCount > 2 else None,
                            ])},

                {'baseAngle':180,
                 'entries':([
                            {'type':bsSpaz.ToughGuyBot,'spacing':6} if playerCount > 3 else None,
                            {'type':bsSpaz.ToughGuyBot,'spacing':6},
                            {'type':bsSpaz.ToughGuyBot,'spacing':6},
                            {'type':bsSpaz.NinjaBot,'spacing':45},
                            {'type':bsSpaz.NinjaBot,'spacing':45} if playerCount > 1 else None,
                            {'type':bsSpaz.ToughGuyBot,'spacing':6},
                            {'type':bsSpaz.ToughGuyBot,'spacing':6},
                            {'type':bsSpaz.ToughGuyBot,'spacing':6} if playerCount > 2 else None,
                            ])},

                # {'baseAngle':90,
                #  'entries':([
                #             {'type':bsSpaz.NinjaBot,'spacing':20},
                #             {'type':bsSpaz.ToughGuyBot,'spacing':10},
                #             {'type':bsSpaz.ToughGuyBot,'spacing':10},
                #             {'type':bsSpaz.ToughGuyBot,'spacing':10} if playerCount > 1 else None,
                #             {'type':bsSpaz.ToughGuyBot,'spacing':10} if playerCount > 3 else None,
                #             {'type':bsSpaz.NinjaBot,'spacing':20},
                #             {'type':None,'spacing':70},
                #             {'type':bsSpaz.BomberBot,'spacing':20},
                #             {'type':bsSpaz.BomberBot,'spacing':20},
                #             {'type':bsSpaz.BomberBot,'spacing':20} if playerCount > 2 else None,
                #             ])},

                {'baseAngle':0,
                 'entries':([
                            {'type':bsSpaz.NinjaBot,'spacing':30},
                            {'type':bsSpaz.ChickBot,'spacing':30},
                            {'type':bsSpaz.ChickBot,'spacing':30},
                            {'type':bsSpaz.ChickBot,'spacing':30},
                            {'type':bsSpaz.ChickBot,'spacing':30} if playerCount > 1 else None,
                            {'type':bsSpaz.ChickBot,'spacing':30} if playerCount > 3 else None,
                            {'type':bsSpaz.NinjaBot,'spacing':30},
                            ])},

                {'baseAngle':90,
                 'entries':([
                            {'type':bsSpaz.MelBot,'spacing':50},
                            {'type':bsSpaz.MelBot,'spacing':50},
                            {'type':bsSpaz.MelBot,'spacing':50},
                            {'type':bsSpaz.MelBot,'spacing':50} if playerCount > 1 else None,
                            {'type':bsSpaz.MelBot,'spacing':50} if playerCount > 3 else None,
                            ])},

                {'baseAngle':0,
                 'entries':([
                            {'type':bsSpaz.ChickBot,'spacing':72},
                            {'type':bsSpaz.ChickBot,'spacing':72},
                            {'type':bsSpaz.ChickBot,'spacing':72},
                            {'type':bsSpaz.ChickBot,'spacing':72},
                            {'type':bsSpaz.ChickBot,'spacing':72},
                            {'type':bsSpaz.ChickBot,'spacing':36} if playerCount > 2 else None,
                        ])},

                {'baseAngle':30,
                 'entries':([
                            {'type':bsSpaz.NinjaBotProShielded,'spacing':50},
                            {'type':bsSpaz.NinjaBotProShielded,'spacing':50},
                            {'type':bsSpaz.NinjaBotProShielded,'spacing':50},
                            {'type':bsSpaz.NinjaBotProShielded,'spacing':50} if playerCount > 1 else None,
                            {'type':bsSpaz.NinjaBotProShielded,'spacing':50} if playerCount > 2 else None,
                            ])},
                ]

        elif self._preset == 'uber':
            self._haveTnt = True
            self._excludePowerups = []
            self._waves = [

                {'entries':[
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopMiddleLeft'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopMiddleRight'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopLeft'} if playerCount > 2 else None,
                        {'type':bsSpaz.PirateBot,'point':'TopRight'},
                        {'type':'delay','duration':4000},
                        {'type':bsSpaz.PirateBot,'point':'TopLeft'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.NinjaBot,'point':'Left'},
                        {'type':bsSpaz.NinjaBot,'point':'Right'},
                        {'type':bsSpaz.NinjaBot,'point':'RightUpperMore'} if playerCount > 2 else None,
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopLeft'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopRight'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.ChickBotPro,'point':'TopRight'},
                        {'type':bsSpaz.ChickBotPro,'point':'RightUpperMore'} if playerCount > 1 else None,
                        {'type':bsSpaz.ChickBotPro,'point':'RightUpper'}, 
                        {'type':bsSpaz.ChickBotPro,'point':'RightLower'},
                        {'type':bsSpaz.ChickBotPro,'point':'RightLowerMore'} if playerCount > 2 else None,
                        {'type':bsSpaz.ChickBotPro,'point':'BottomRight'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.NinjaBotProShielded,'point':'BottomRight'},
                        {'type':bsSpaz.NinjaBotProShielded,'point':'Bottom'} if playerCount > 2 else None,
                        {'type':bsSpaz.NinjaBotProShielded,'point':'BottomLeft'},
                        {'type':bsSpaz.NinjaBotProShielded,'point':'Top'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopMiddle'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.PirateBot,'point':'LeftUpper'},
                        {'type':'delay','duration':1000},
                        {'type':bsSpaz.ToughGuyBotProShielded,'point':'LeftLower'},
                        {'type':bsSpaz.ToughGuyBotProShielded,'point':'LeftLowerMore'},
                        {'type':'delay','duration':4000},
                        {'type':bsSpaz.PirateBot,'point':'RightUpper'},
                        {'type':'delay','duration':1000},
                        {'type':bsSpaz.ToughGuyBotProShielded,'point':'RightLower'},
                        {'type':bsSpaz.ToughGuyBotProShielded,'point':'RightUpperMore'},
                        {'type':'delay','duration':4000},
                        {'type':bsSpaz.PirateBot,'point':'Left'},
                        {'type':'delay','duration':5000},
                        {'type':bsSpaz.PirateBot,'point':'Right'},
                        ]},

                {'entries':[
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopLeft'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopRight'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretBottomLeft'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretBottomRight'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopMiddleLeft'},
                        {'type':bsSpaz.BomberBotProStatic,'point':'TurretTopMiddleRight'},
                        ]},
                ]
        # we generate these on the fly in endless..
        elif self._preset == 'endless':
            self.createControlsHelpUI(durationScale=0.5)
            self._haveTnt = True
            self._excludePowerups = []
        
        else: raise Exception("Invalid preset: "+str(self._preset))

        # spit out a few powerups and start dropping more shortly
        self._dropPowerups(standardPoints=True,powerupType='curse' if self._preset == 'uber' else ('landMines' if self._preset == 'rookie' else None))
        bs.gameTimer(4000,bs.WeakCall(self._startPowerupDrops))
        self.createLifeWarningTimer()
        self._updateScores()
        self._bots = bsSpaz.BotSet(self.players)

        # our TNT spawner (if applicable)
        if self._haveTnt:
            self._tntSpawner = bsBomb.TNTSpawner(activity=self,position=self._tntSpawnPosition)

        bs.gameTimer(4000,bs.WeakCall(self._startUpdatingWaves))

    def onGotScoresToBeat(self,scores):
        self.showStandardScoresToBeatUI(scores)

    def spawnPlayer(self,player):

        # we keep track of who got hurt each wave for score purposes
        player.data['hasBeenHurt'] = False
    
        pos = (self._spawnCenter[0]+random.uniform(-1.5,1.5),self._spawnCenter[1],self._spawnCenter[2]+random.uniform(-1.5,1.5))
        s = self.spawnSpazForPlayer(player,position=pos)
        s.dropBombCallback = bs.WeakCall(self._handlePlayerDroppedBomb)

        # add the material that causes us to hit the player-wall (on courtyard)
        if self._levelType == bsMap.CourtyardMap:
            for attr in ['materials','rollerMaterials']:
                materials = getattr(s.node,attr)
                setattr(s.node,attr,materials + (self._level.collideWithWallMaterial,))

    def _handlePlayerDroppedBomb(self,player):
        self._playerHasDroppedBomb = True

    def _dropPowerup(self,index,powerupType=None):
        powerupType = bsPowerup.getFactory().getRandomPowerupType(forceType=powerupType,excludeTypes=self._excludePowerups)
        bsPowerup.Powerup(self,bs.Actor.RETAINED,translate=self._level.powerupSpawnPoints[index],powerupType=powerupType)

    def _startPowerupDrops(self):
        self._powerupDropTimer = bs.Timer(3000,bs.WeakCall(self._dropPowerups),repeat=True)

    def _dropPowerups(self,standardPoints=False,powerupType=None):
        """ Generic powerup drop """

        if standardPoints:
            pts = self._level.powerupSpawnPoints
            for i,pt in enumerate(pts):
                bs.gameTimer(1000+i*500,bs.WeakCall(self._dropPowerup,i,powerupType if i == 0 else None))
        else:
            pt = (self._powerupCenter[0]+random.uniform(-1.0*self._powerupSpread[0],1.0*self._powerupSpread[0]),
                  self._powerupCenter[1],self._powerupCenter[2]+random.uniform(-self._powerupSpread[1],self._powerupSpread[1]))

            # drop one random one somewhere..
            bsPowerup.Powerup(self,bs.Actor.RETAINED,translate=pt,powerupType=bsPowerup.getFactory().getRandomPowerupType(excludeTypes=self._excludePowerups))

    def doEnd(self,outcome,delay=0):

        if outcome == 'defeat':
            self.fadeToRed()

        if self._wave >= 2:
            score = self._score
            failMessage = None
        else:
            score = None
            failMessage = bs.getResource('reachWave2Text')
        self.end({'outcome':outcome,'score':score,'failMessage':failMessage,'playerInfo':self.initialPlayerInfo},delay=delay)

    def _updateWaves(self):

        # if we have no living bots, go to the next wave
        if self._canEndWave and not self._bots.haveLivingBots() and not self._gameOver:

            self._canEndWave = False

            self._timeBonusTimer = None
            self._timeBonusText = None

            if self._preset == 'endless': won = False
            else: won = (self._wave == len(self._waves))

            # reward time bonus
            baseDelay = 4000 if won else 0

            if self._timeBonus > 0:
                bs.gameTimer(0,bs.WeakCall(bs.playSound,self._cashRegisterSound))
                bs.gameTimer(baseDelay,bs.WeakCall(self._awardTimeBonus,self._timeBonus))
                baseDelay += 1000

            # reward flawless bonus
            if self._wave > 0:
                haveFlawless = False
                for player in self.players.values():
                    if not player.actor.dead and player.data['hasBeenHurt'] == False:
                        haveFlawless = True
                        bs.gameTimer(baseDelay,bs.WeakCall(self._awardFlawlessBonus,player))
                    player.data['hasBeenHurt'] = False # reset
                if haveFlawless: baseDelay += 1000

            if won:

                # give remaining players some points and have them celebrate
                self._victoryText = bsAnnounce.ZoomText(self,bs.Actor.UNRETAINED,
                                                        bs.getResource('victoryText'),maxWidth=800,
                                                        lifespan=4000,jitter=2.0,position=(0,200))
                self.celebrate(20000)

                # rookie onslaught completion
                if self._preset == 'training':
                    self.awardAchievement('Onslaught Training Victory',sound=False)
                    if not self._playerHasDroppedBomb:
                        self.awardAchievement('Boxer',sound=False)
                elif self._preset == 'rookie':
                    self.awardAchievement('Rookie Onslaught Victory',sound=False)
                    if not self._aPlayerHasBeenHurt:
                        self.awardAchievement('Flawless Victory',sound=False)
                elif self._preset == 'pro':
                    self.awardAchievement('Pro Onslaught Victory',sound=False)
                    if not self._playerHasDroppedBomb:
                        self.awardAchievement('Pro Boxer',sound=False)
                elif self._preset == 'uber':
                    self.awardAchievement('Uber Onslaught Victory',sound=False)

                bs.gameTimer(baseDelay,bs.WeakCall(self._awardCompletionBonus))
                baseDelay += 850
                bs.playSound(self._winSound)
                self.cameraFlash()
                bsUtils.playMusic('Victory')
                self._gameOver = True

                # cant just pass delay to doEnd because our extra bonuses havnt been added yet
                # (once we call doEnd the score gets locked in)
                bs.gameTimer(baseDelay,bs.WeakCall(self.doEnd,'victory'))

                return

            self._wave += 1

            # short celebration after waves
            if self._wave > 1: self.celebrate(500)
                            
            bs.gameTimer(baseDelay,bs.WeakCall(self._startNextWave))

    def _awardCompletionBonus(self):
        bs.playSound(self._cashRegisterSound)
        for player in self.players.values():
            try:
                if not player.actor.dead:
                    self.scoreSet.playerScored(player,int(100/len(self.initialPlayerInfo)),scale=1.4,color=(0.6,0.6,1.0,1.0),title=bs.getResource('completionBonusText'),screenMessage=False)
            except Exception: pass
        
    def _awardTimeBonus(self,bonus):
        bs.playSound(self._cashRegisterSound)
        bsAnnounce.AchievementText(self,bs.Actor.RETAINED,
                                   '+'+str(bonus)+' '+bs.getResource('timeBonusText'),
                                   color=(1,1,0.5,1),
                                   scale=1.0,
                                   position=(0,3,-1))
        self._score += self._timeBonus
        self._updateScores()

    def _awardFlawlessBonus(self,player):
        bs.playSound(self._cashRegisterSound)
        try:
            if not player.actor.dead:
                self.scoreSet.playerScored(player,self._flawlessBonus,scale=1.2,color=(0.6,1.0,0.6,1.0),title=bs.getResource('flawlessWaveText'),screenMessage=False)
        except Exception: pass


    def _startTimeBonusTimer(self):
        self._timeBonusTimer = bs.Timer(1000,bs.WeakCall(self._updateTimeBonus),repeat=True)
        
    def _updatePlayerSpawnInfo(self):

        # if we have no living teams lets just blank this
        if (len(self.getLivingTeams()) == 0):
            self._spawnInfoText.node.text = ''
        else:
            t = ''
            for player in self.players.values():
                try: exists = not player.actor.dead
                except Exception: exists = False
                if not exists and (self._preset == 'endless' or (player.data['respawnWave'] <= len(self._waves))):
                    t += bs.getResource('onslaughtRespawnText').replace('${PLAYER}',bs.getPlayerName(player.playerID)).replace('${WAVE}',str(player.data['respawnWave']))+'\n'
            self._spawnInfoText.node.text = t

    def _startNextWave(self):

        # this could happen if we beat a wave as we die..
        if self._gameOver: return

        # respawn applicable players
        if self._wave > 1:
            for player in self.players.values():
                try: exists = not player.actor.dead
                except Exception: exists = False
                if not exists and player.data['respawnWave'] == self._wave:
                    self.spawnPlayer(player)

        self._updatePlayerSpawnInfo()

        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,bs.getResource('waveText')+' '+str(self._wave),
                            maxWidth=800,lifespan=1000,jitter=2.0,position=(0,200))
        bs.gameTimer(400,bs.WeakCall(bs.playSound,self._newWaveSound))

        t = 0
        dt = 200
        botAngle = random.random()*360.0

        if self._wave == 1:
            spawnTime = 3973
            t += 500
        else:
            spawnTime = 2648


        offs = 0 # debugging

        # populate waves

        # generate random waves in endless mode
        if self._preset == 'endless':

            level = self._wave

            botTypes2 = [bsSpaz.BomberBot,
                         bsSpaz.ToughGuyBot,
                         bsSpaz.ChickBot,
                         bsSpaz.NinjaBot,
                         bsSpaz.BomberBotPro,
                         bsSpaz.ToughGuyBotPro,
                         bsSpaz.ChickBotPro,
                         bsSpaz.BomberBotProShielded,
                         bsSpaz.PirateBot,
                         bsSpaz.NinjaBotProShielded,
                         bsSpaz.MelBot,
                         bsSpaz.ToughGuyBotProShielded,
                         bsSpaz.ChickBotProShielded
                         ]

            if level > 5:
                botTypes2 += [
                         bsSpaz.PirateBot,
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ToughGuyBotProShielded,
                         bsSpaz.NinjaBotProShielded,
                    ]
            if level > 7:
                botTypes2 += [
                         bsSpaz.PirateBot,
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ToughGuyBotProShielded,
                         bsSpaz.NinjaBotProShielded,
                    ]
            if level > 10:
                botTypes2 += [
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ChickBotProShielded
                    ]
            if level > 13:
                botTypes2 += [
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ChickBotProShielded,
                         bsSpaz.ChickBotProShielded
                    ]

            botLevels = [[b for b in botTypes2 if b.pointsMult == 1],
                         [b for b in botTypes2 if b.pointsMult == 2],
                         [b for b in botTypes2 if b.pointsMult == 3],
                         [b for b in botTypes2 if b.pointsMult == 4]]
            if not all([len(a) > 0 for a in botLevels]): raise Exception()

            targetPoints = level*3-2
            minDudes = min(1+level/3,10)
            maxDudes = min(10,level+1)
            maxLevel = 4 if level > 6 else (3 if level > 3 else (2 if level > 2 else 1))
            groupCount = 3
            distribution = _getDistribution(targetPoints,minDudes,maxDudes,groupCount,maxLevel)

            allEntries = []
            for group in distribution:
                entries = []
                for entry in group:
                    botLevel = botLevels[entry[0]-1]
                    botType = botLevel[random.randrange(len(botLevel))]
                    r = random.random()
                    if r < 0.5: spacing = 10
                    elif r < 0.9: spacing = 20
                    else: spacing = 40
                    split = random.random() > 0.3
                    for i in range(entry[1]):
                        if split and i%2==0: entries.insert(0,{"type":botType, "spacing":spacing})
                        else: entries.append({"type":botType, "spacing":spacing})
                if len(entries) > 0:
                    allEntries += entries
                    allEntries.append({"type":None,"spacing":40 if random.random() < 0.5 else 80})

            angleRand = random.random()
            if angleRand > 0.75: baseAngle = 130
            elif angleRand > 0.5: baseAngle = 210
            elif angleRand > 0.25: baseAngle = 20
            else: baseAngle = -30
            baseAngle += (0.5-random.random())*20.0

            wave = {'baseAngle':baseAngle,
                    'entries':allEntries}
        else:
            wave = self._waves[self._wave-1]


        entries = []

        try: botAngle = wave['baseAngle']
        except Exception: botAngle = 0

        entries += wave['entries']

        thisTimeBonus = 0
        thisFlawlessBonus = 0

        for info in entries:
            if info == None: continue

            botType = info['type']

            if botType == 'delay':
                spawnTime += info['duration']
                continue
            if botType != None:
                thisTimeBonus += botType.pointsMult * 20
                thisFlawlessBonus += botType.pointsMult * 5
            # if its got a position, use that
            try: point = info['point']
            except Exception: point = None
            if point != None:
                bs.gameTimer(t,bs.WeakCall(self.addBotAtPoint,point,botType,spawnTime))
                t += dt
            else:
                try: spacing = info['spacing']
                except Exception: spacing = 5.0
                botAngle += spacing*0.5
                if botType != None:
                    bs.gameTimer(t,bs.WeakCall(self.addBotAtAngle,botAngle,botType,spawnTime))
                    t += dt
                botAngle += spacing*0.5
            
        # we can end the wave after all the spawning happens
        bs.gameTimer(t+spawnTime-dt+10,bs.WeakCall(self._setCanEndWave))

        # reset our time bonus
        self._timeBonus = thisTimeBonus
        self._flawlessBonus = thisFlawlessBonus
        self._timeBonusText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                           'text',{"vAttach":'top',
                                                   "hAttach":'center',
                                                   "hAlign":'center',
                                                   "color":(1,1,0.5,1),
                                                   "translate":(0,-60),
                                                   "scale":[0.6],
                                                   "text":bs.getResource('timeBonusText')+': '+str(self._timeBonus)},
                                           name="Time Bonus Left")
        bs.gameTimer(5000,bs.WeakCall(self._startTimeBonusTimer))
        self._waveText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                      'text',{"vAttach":'top',
                                              "hAttach":'center',
                                              "hAlign":'center',
                                              "color":(0.6,0.6,0.6,1.0),
                                              "translate":(0,-40),
                                              "scale":[1.1],
                                              "text":bs.getResource('waveText')+' '+str(self._wave)+('' if self._preset == 'endless' else ('/'+str(len(self._waves))))},
                                      name='Wave Number')

    def addBotAtPoint(self,point,spazType,spawnTime=1000):
        # dont add if the game has ended
        if self._gameOver: return
        pt = self._level.defs.points['botSpawn'+point]
        bsUtils.Spawner(activity=self,data=spazType,pt=pt,spawnTime=spawnTime)
        
    def addBotAtAngle(self,angle,spazType,spawnTime=1000):

        # dont add if the game has ended
        if self._gameOver: return

        angleRadians = angle/57.2957795
        x = math.sin(angleRadians)*1.06
        z = math.cos(angleRadians)*1.06

        pt = (x/0.125,2.3,(z/0.2)-3.7)

        bsUtils.Spawner(activity=self,data=spazType,pt=pt,spawnTime=spawnTime)

    def _updateTimeBonus(self):
        self._timeBonus = int(self._timeBonus * 0.93)
        if self._timeBonus > 0 and self._timeBonusText != None:
            self._timeBonusText.node.text = bs.getResource('timeBonusText')+': '+str(self._timeBonus)
        else: self._timeBonusText = None

    def _startUpdatingWaves(self):
        self._waveUpdateTimer = bs.Timer(2000,bs.WeakCall(self._updateWaves),repeat=True)
        
    def _updateScores(self):

        if self._preset == 'endless':
            if self._score >= 500: self.awardAchievement('Onslaught Master')
            if self._score >= 1000: self.awardAchievement('Onslaught Wizard')
            if self._score >= 5000: self.awardAchievement('Onslaught God')

        self._scoreBoard.setTeamValue(self.teams[0],self._score,maxScore=None)

        
    def handleMessage(self,m):

        if isinstance(m,bsSpaz.PlayerSpazHurtMessage):
            player = m.spaz.getPlayer()
            if player is None: return
            player.data['hasBeenHurt'] = True
            self._aPlayerHasBeenHurt = True

        elif isinstance(m,bsScoreSet.PlayerScoredMessage):
            self._score += m.score
            self._updateScores()

        elif isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            self.__superHandleMessage(m) # augment standard behavior
            player = m.spaz.getPlayer()
            if player is None: return
            self._aPlayerHasBeenHurt = True
            # make note with the player when they can respawn
            if self._wave < 10: player.data['respawnWave'] = max(2,self._wave+1)
            elif self._wave < 15: player.data['respawnWave'] = max(2,self._wave+2)
            else: player.data['respawnWave'] = max(2,self._wave+3)
            bs.gameTimer(100,self._updatePlayerSpawnInfo)
            bs.gameTimer(100,self._checkRoundOver)

        elif isinstance(m,bsSpaz.SpazBotDeathMessage):
            pts,importance = m.badGuy.getDeathPoints(m.how)
            if m.killer is not None:

                # toss-off-map achievement
                if self._preset == 'training':
                    if m.badGuy.lastAttackedType == ('pickedUp','default'):
                        if not hasattr(self,'_throwOffKills'): self._throwOffKills = 0
                        self._throwOffKills += 1
                        if self._throwOffKills >= 3: 
                            self.awardAchievement('Off You Go Then')

                # land-mine achievement
                elif self._preset == 'rookie':
                    if m.badGuy.lastAttackedType == ('explosion','landMine'):
                        if not hasattr(self,'_landMineKills'): self._landMineKills = 0
                        self._landMineKills += 1
                        if self._landMineKills >= 3: 
                            self.awardAchievement('Mine Games')

                # tnt achievement
                elif self._preset == 'pro':
                    if m.badGuy.lastAttackedType == ('explosion','tnt'):
                        if not hasattr(self,'_tntKills'): self._tntKills = 0
                        self._tntKills += 1
                        if self._tntKills >= 3: 
                            bs.gameTimer(500,bs.WeakCall(self.awardAchievement,'Boom Goes the Dynamite'))

                elif self._preset == 'uber':

                    # uber mine achievement
                    if m.badGuy.lastAttackedType == ('explosion','landMine'):
                        if not hasattr(self,'_landMineKills'): self._landMineKills = 0
                        self._landMineKills += 1
                        if self._landMineKills >= 6: 
                            self.awardAchievement('Gold Miner')
                    # uber tnt achievement
                    if m.badGuy.lastAttackedType == ('explosion','tnt'):
                        if not hasattr(self,'_tntKills'): self._tntKills = 0
                        self._tntKills += 1
                        if self._tntKills >= 6: 
                            bs.gameTimer(500,bs.WeakCall(self.awardAchievement,'TNT Terror'))

                try: target = m.badGuy.node.translate
                except Exception: target = None
                try:
                    killerPlayer = self.players[m.killer]
                    self.scoreSet.playerScored(killerPlayer,pts,target=target,kill=True,screenMessage=False,importance=importance)
                    bs.playSound(self._dingSound if importance == 1 else self._dingSoundHigh,volume=0.6)
                except Exception: pass
            # normally we pull scores from the score-set, but if there's no player lets be explicit..
            else: self._score += pts
            self._updateScores()
        elif isinstance(m,bsUtils.Spawner.SpawnMessage):
            spazType = m.data
            spaz = spazType(self,bs.Actor.UNRETAINED,level=self._level)
            bs.playSound(self._spawnSound,position=m.pt)
            bs.nodeMessage(spaz.node,"flash")
            spaz.node.isAreaOfInterest = 0
            spaz.handleMessage(bs.StandMessage(m.pt,random.uniform(0,360)))
            self._bots.addBot(spaz)
        else:
            self.__superHandleMessage(m)

    def _setCanEndWave(self):
        self._canEndWave = True

    def __superHandleMessage(self,m):
        super(OnslaughtGame,self).handleMessage(m)


    def _checkRoundOver(self):
        """ see if the round is over in response to an event (player died, etc) """

        # if we already ended it doesn't matter
        if self.hasEnded(): return

        livingTeams = self.getLivingTeams()
        if (len(livingTeams) == 0):

            # tell our bots to celebrate just to rub it in
            self._bots.finalCelebrate()
                    
            self._gameOver = True
            self.doEnd('defeat',delay=2000)
            bsUtils.playMusic(None)


